using System;
using System.IO;
using UnityEngine;
using UnityEngine.UI;
using AssetBundles;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using MeshVR;
using MVR.FileManagementSecure;
using AssetBundles;
using SimpleJSON;
using System.Text.RegularExpressions;
using Battlehub.RTSaveLoad;
using Battlehub.RTCommon;

namespace Stopper {
	public class AlternativeFuta : MVRScript {
	
	string packageId;
	string packagePath;
	
	string altFutaSavePath = "Saves\\AltFuta";
	string altFutaFileExt = "altfuta";
	
	// Character skin prefabs
	GameObject femaleSkinsPrefab;
	GameObject maleSkinsPrefab;
	
	GameObject femaleSkinsContainer;

	GameObject maleAsset;
	
	// Stuff for the new 'DAZCharacter'
	// that has our new modified skin
	GameObject newCharacterContainer;
	DAZCharacter newCharacter;
	
	DAZCharacterSelector characterSelector;
	
	DAZCharacterRun characterRun;
	
	DAZMergedSkinV2 skin;
		
	DAZMergedMesh mergedMesh;
	
	DAZMesh genMesh;
	
	GameObject femaleSkinBase;
	
	DAZBone gen1Bone;
	DAZBone gen2Bone;
	DAZBone gen3Bone;
	
	string futaSkinName = "AltFutaSkin";
	
	
	// Previous
	DAZCharacter previousSelected;
	
	
	Dictionary<string, string> femaleSkinsWithMatchingFuta = new Dictionary<string, string>()
	{
		{ "renkayla", "FutaKayla" },
		{ "renlexi", "FutaLexi" },
		{ "renmaria", "FutaMaria" },
		{ "renmia", "FutaMia" },
		{ "rensimone", "FutaSimone" },
	};
	
	// We also instantiate a male character to steal
	// some data from during initialization
	GameObject maleCharacterContainer;
	GameObject maleSkinBase;
	DAZMergedSkinV2 maleSkin;
	DAZMergedMesh maleMergedMesh;
	
	List<GameObject> maleObjects = new List<GameObject>();
	
	// UI Stuff
	GameObject UIContainer;
	
	GameObject materialUIPrefab;
	MVRScriptUI scriptUI;
	
	JSONStorableBool hideFemaleGens;
	JSONStorableBool useTransparentShader;
	JSONStorableBool hideMalePelvis;
	JSONStorableBool mergeLoad;
	
	bool skinLoaded = false;
	bool useVaginaAsPelvisBase = false;
	bool overrideFemalePelvisTextures = false;
	
	// Mat Options
	
	MaterialOptionsWrapper genMaterialOptions;
	Transform genMaterialOptionsUI;
	
	//MaterialOptionsWrapper pelvisMaterialOptions;
	//Transform pelvisMaterialOptionsUI;
	
	//MaterialOptionsWrapper outerPelvisMaterialOptions;
	//Transform outerPelvisMaterialOptionsUI;
	
	// Morph Stuff
	//float genBoneYOffsetDefault = 0.028f;
	//float genBoneZOffsetDefault = -0.016f
	
	float genBoneYOffsetDefault = 0.030f;
	float genBoneZOffsetDefault = -0.010f;
	float hidePerineumDefault = 0f;// 0.5f;
	
	DAZMorph genYOffsetMorph;
	DAZMorph genZOffsetMorph;
	DAZMorph genYMeshOffsetMorph;
	DAZMorph genZMeshOffsetMorph;
	DAZMorph hidePerineumMorph;
	DAZMorph gapFixMorph;
	DAZMorph vaginaHideMorph;
	
	JSONStorableFloat genYOffsetValue;
	JSONStorableFloat genZOffsetValue;
	JSONStorableFloat genYMeshOffsetValue;
	JSONStorableFloat genZMeshOffsetValue;
	JSONStorableFloat hidePerineumOffsetValue;
	JSONStorableFloat gapFixOffsetValue;
	JSONStorableFloat vaginaHideOffsetValue;
	
	
	// Saving and loading
	JSONClass savedPresetJson;
	
	JSONClass[] oldSkinOpts;
	JSONClass oldTextureControlJson;
	
	// Physics stuff
	List<string> genColliderIgnoreNames = new List<string>{
			// Front pelvis
			"AutoColliderpelvisF1","AutoColliderpelvisF2","AutoColliderpelvisF3",
			"AutoColliderpelvisF4","AutoColliderpelvisF5","AutoColliderpelvisF6",
			// Inner pelvis
			"AutoColliderpelvisFL1","AutoColliderpelvisFL1b", "AutoColliderpelvisFL2","AutoColliderpelvisFL3","AutoColliderpelvisFL4","AutoColliderpelvisFL5",
			"AutoColliderpelvisFR1","AutoColliderpelvisFR1b", "AutoColliderpelvisFR2","AutoColliderpelvisFR3","AutoColliderpelvisFR4","AutoColliderpelvisFR5",
			// Vagina
			"JointG", "Collider1"
	};
	List<AutoCollider> genAutoColliderIgnoreList = new List<AutoCollider>();
	List<Collider> genColliderIgnoreList = new List<Collider>();
	
	HashSet<AutoCollider> genAutoColliders = new HashSet<AutoCollider>();
	HashSet<Collider> genColliders = new HashSet<Collider>();
	
	Dictionary<AutoCollider, AutoColliderConfig> maleAutoColliderConfigs = new Dictionary<AutoCollider, AutoColliderConfig>();
	
	public class AutoColliderConfig {
		public int targetVertex = -1;
		public int anchorVertex1 = -1;
		public int anchorVertex2 = -1;
		public int oppositeVertex = -1;
	}
	
	// Small wrappers that help us get access to some protected members
	public class DAZSkinV2Wrapper : DAZSkinV2 {

		public DAZSkinV2Wrapper() {
			this.GPUmaterials = new Material[0];
		}

		public void Copy(DAZSkinV2 other){
			this.sceneGeometryId = other.sceneGeometryId;
			this._numBones = other.numBones;
			this.nodes = other.nodes;
			this.dazMesh = other.dazMesh;
		}
	}
	
	// This is used to detect when the selected character is changed
	public class DisableListener : MonoBehaviour {
		public AlternativeFuta script;
		
		public void OnDisable(){
			if(script == null) return;
			
			
		
			// Restore modified auto colliders
			if(script.maleAutoColliderConfigs != null){
				foreach(var entry in script.maleAutoColliderConfigs){
					AutoCollider ac = entry.Key;
					AutoColliderConfig config = entry.Value;
					
					ac.targetVertex = config.targetVertex;
					ac.anchorVertex1 = config.anchorVertex1;
					ac.anchorVertex2 = config.anchorVertex2;
					ac.oppositeVertex = config.oppositeVertex;
					
					ac.AutoColliderSizeSet(true);
				}
			}
		}
	}

	public class MaterialOptionsWrapper : DAZCharacterMaterialOptions {
		public void SetParams(){
			this.SetAllParameters();
		}
		
		public void CopyMaterialParams(DAZCharacterMaterialOptions src){
			if(color1JSONParam != null) color1JSONParam.SetColor(src.color1CurrentColor);
			if(color2JSONParam != null) color2JSONParam.SetColor(src.color2CurrentColor);
			if(color3JSONParam != null) color3JSONParam.SetColor(src.color3CurrentColor);
			
			if(param1JSONParam != null) param1JSONParam.val = src.param1CurrentValue;
			if(param2JSONParam != null) param2JSONParam.val = src.param2CurrentValue;
			if(param3JSONParam != null) param3JSONParam.val = src.param3CurrentValue;
			if(param4JSONParam != null) param4JSONParam.val = src.param4CurrentValue;
			if(param5JSONParam != null) param5JSONParam.val = src.param5CurrentValue;
			
			// Param indexes don't match past this point
			
			if(param6JSONParam != null) param6JSONParam.val = src.param9CurrentValue;
			// Param 7 is Alpha adjust 
			if(param8JSONParam != null) param8JSONParam.val = src.param6CurrentValue;
			if(param9JSONParam != null) param9JSONParam.val = src.param7CurrentValue;
			if(param10JSONParam != null) param10JSONParam.val = src.param8CurrentValue;
			
			SetParams();
		}
		
		public void ClearCustomTextures(){
			if(customTexture1UrlJSON != null) customTexture1UrlJSON.Clear();
			if(customTexture2UrlJSON != null) customTexture2UrlJSON.Clear();
			if(customTexture3UrlJSON != null) customTexture3UrlJSON.Clear();
			if(customTexture4UrlJSON != null) customTexture4UrlJSON.Clear();
			if(customTexture5UrlJSON != null) customTexture5UrlJSON.Clear();
			if(customTexture6UrlJSON != null) customTexture6UrlJSON.Clear();
		}
	}
	
	public override void Init() {
		packageId = GetPackageId(this);
		packagePath = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";
	
	
		// Initialize some references. 		
		var geometry = containingAtom.GetStorableByID("geometry");
		characterRun = geometry.GetComponent<DAZCharacterRun>();
		
		gen1Bone = characterRun.bones.GetDAZBone("Gen1");
		gen2Bone = characterRun.bones.GetDAZBone("Gen2");
		gen3Bone = characterRun.bones.GetDAZBone("Gen3");
		
		GetMaleTransforms();
		SaveAutoColliderConfigs();
		
		//SetMaleTransformsActive(true);
		
		characterSelector = geometry.GetComponent<DAZCharacterSelector>();
		previousSelected = characterSelector.selectedCharacter;
		
		// TODO: Replace these with hard-coded paths
		foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
			if(go.name =="FemaleCharactersPrefab"){
				femaleSkinsPrefab = go;
			}
			
			if(go.name =="MaleCharactersPrefab"){
				maleSkinsPrefab = go;
			}
		}
		
		foreach(GameObject go in GetAllChildren(geometry.gameObject)){
			if(go.name.Contains("FemaleCharactersPrefab")){
				femaleSkinsContainer = go;
			}
		}
		
		// Store previous material options
		try {		
			if(previousSelected != null){
				DAZCharacterMaterialOptions[] oldOptions = previousSelected.GetComponentsInChildren<DAZCharacterMaterialOptions>();
				oldSkinOpts = new JSONClass[oldOptions.Length];
				for(int j = 0; j < oldOptions.Length; j++){
					var json = oldOptions[j].GetJSON(true,true,true);
					oldSkinOpts[j] = json;
				}
			}
		} catch (Exception e) {
			SuperController.LogError($"Error storing previous skin options. {e}");
		}
		
		// Store other material settings in DAZCharacterTextureControl
		var oldTextureControl = previousSelected.GetComponentInChildren<DAZCharacterTextureControl>();
		if(oldTextureControl != null){
			oldTextureControlJson = oldTextureControl.GetJSON(true,true,true);
		}
		
		
		// Select the base skin.
		// We'll try to use whatever selected, otherwise we choose 'rensimone' arbitrarily
		string baseSkinName = null;
		if(characterSelector.selectedCharacter != null) {
			baseSkinName = characterSelector.selectedCharacter.name;
			
		}
		
		if(baseSkinName == null || baseSkinName == futaSkinName){
			baseSkinName = "rensimone";
		}
		
		/*foreach(GameObject go in GetAllChildren(femaleSkinsPrefab)){
			if(go.name == baseSkinName){
				femaleSkinBase = go;
			}
		}*/

		femaleSkinBase = previousSelected.gameObject;
		
		if(femaleSkinBase == null){
			SuperController.LogMessage($"Unable to find female skin preset for '{baseSkinName}'");
			return;
		}
		
		
		string maleSkinName = null;
		femaleSkinsWithMatchingFuta.TryGetValue(baseSkinName, out maleSkinName);
		if(maleSkinName == null){
			maleSkinName = "FutaSimone";
		}
		
		foreach(GameObject go in GetAllChildren(maleSkinsPrefab)){
			if(go.name == maleSkinName){
				maleSkinBase = go;
			}
		}
		
		var physicsModel = containingAtom.GetStorableByID("PhysicsModel");
		foreach(AutoCollider ac in physicsModel.GetComponentsInChildren<AutoCollider>()){
			foreach(string toIgnore in genColliderIgnoreNames){
				if(ac.name.Contains(toIgnore) || ac.gameObject.transform.parent.name.Contains(toIgnore) ){
					genAutoColliderIgnoreList.Add(ac);
				}
			}
		}
		
		foreach(Collider c in physicsModel.GetComponentsInChildren<Collider>()){
			foreach(string toIgnore in genColliderIgnoreNames){
				if(c.name.Contains(toIgnore) || c.gameObject.transform.parent.name.Contains(toIgnore)){
					genColliderIgnoreList.Add(c);
				}
			}
		}

		InitMainUI();

		AddMaleGenitalia();
	}
	
	
	private void GetMaleTransforms(){
		// TODO: Use hardcoded paths for these
		var physicsModel = containingAtom.GetStorableByID("PhysicsModel");
		
		foreach(GameObject go in GetAllChildren(physicsModel.gameObject)){
			if( go.name == "Gen1"
				|| go.name == "Gen2"
				|| go.name == "Gen3"
				|| go.name == "Testes"){
				maleObjects.Add(go);
				
				foreach(var ac in go.GetComponentsInChildren<AutoCollider>(true)){
					genAutoColliders.Add(ac);
				}
				
				foreach(var c in go.GetComponentsInChildren<Collider>(true)){
					genColliders.Add(c);
				}
				
			}
		}
		maleObjects.Add(containingAtom.GetStorableByID("penisTipControl").gameObject);
		maleObjects.Add(containingAtom.GetStorableByID("penisMidControl").gameObject);
		maleObjects.Add(containingAtom.GetStorableByID("penisBaseControl").gameObject);
		maleObjects.Add(containingAtom.GetStorableByID("testesControl").gameObject);
	}
	
	private void SaveAutoColliderConfigs(){
		foreach(AutoCollider ac in genAutoColliders){
			maleAutoColliderConfigs[ac] = new AutoColliderConfig{ 
				targetVertex = ac.targetVertex,
				anchorVertex1 = ac.anchorVertex1,
				anchorVertex2 = ac.anchorVertex2,
				oppositeVertex = ac.oppositeVertex,
			};
		}
	}
	
	public void Start() {
		StartCoroutine("WaitForLoad");
		StartCoroutine("InitMaterialUI");
	}
	
	private void ReloadSkin(){
		if(skinLoaded){
			skinLoaded = false;
			savedPresetJson = GetJSON(true,true,true);
			RestoreFromJSON(savedPresetJson);
			
			UnloadSkin();
			
			AddMaleGenitalia();
			StartCoroutine("WaitForLoad");
			StartCoroutine("InitMaterialUI");
		}
	}

	private GameObject LoadAssetBundleSync(string assetBundleName, string assetName) {
		GameObject asset = null;
		
		string error;
		var loadedAssetBundle = AssetBundleManager.GetLoadedAssetBundle(assetBundleName, out error);

		if(loadedAssetBundle == null) {
			AssetBundleLoader loader = new AssetBundleLoader();
			loader.Load(assetBundleName, (bundleName, assetBundle) => {
				var loadedAsset = assetBundle.LoadAsset(assetName, typeof(GameObject)) as GameObject;
				asset = UnityEngine.Object.Instantiate<GameObject>(loadedAsset);
				
				assetBundle.Unload(true);
			});
		} else {
			asset = UnityEngine.Object.Instantiate<GameObject>(loadedAssetBundle.m_AssetBundle.LoadAsset(assetName, typeof(GameObject)) as GameObject);
		}
		
		return asset;
	}

	private void AddMaleGenitalia() {
		
		SetMaleTransformsActive(true);
		/* 	The VAM skin hierarchy includes a DAZMergedSkinV2, which uses a DAZMergedMesh. 
			'Merged' here refers to the fact that genitals have their own DAZMesh that is 
			grafted onto the main mesh. There is code for 2 graft meshes and, luckily, the 
			female model only uses one. 
		
			The first thing we'll do is set up a new skin. This requires a few high-level steps:
			1. Create a new DAZCharacter off of a prefab
				- This will create most of our DAZMergedMesh and our DAZMergedSkinV2
			2. Add a DAZMesh and DAZSkinV2 for the penis prior to the main skin initializing
				- The DAZMesh we get from a file
				- The DAZSkinV2 we'll copy from a male skin
			3. Merge the DAZMergedMesh and our DAZMergedSkinV2
			4. Set the materials
		*/
		
		// Initialize the male skin that we'll steal some data from.
		// Specifically, we want the node weights, which determine how 
		// each vertex is weiged to the DAZBones of the skeleton
		maleCharacterContainer = UnityEngine.Object.Instantiate<GameObject>(maleSkinBase);
		
		// Enable the object. This will cause the DAZCharacter to load skin
		// artifacts from assetBundles
		maleCharacterContainer.SetActive(true);
		
		DAZCharacter maleCharacter = maleCharacterContainer.GetComponentInChildren<DAZCharacter>(true);
		
		if(maleAsset == null) {
			maleAsset = LoadAssetBundleSync(maleCharacter.assetBundleName, maleCharacter.assetName);
			maleAsset.SetActive(true);
		}
		
		
		DAZMergedSkinV2 maleMergedSkin = maleAsset.GetComponentInChildren<DAZMergedSkinV2>(true);
		maleMergedSkin.enabled = true;
		maleMergedSkin.CopyMaterials();

		maleMergedMesh = (maleMergedSkin.dazMesh as DAZMergedMesh);
			
		// Initialize the new skin, which will initially include just the DAZCharacter
		newCharacterContainer = UnityEngine.Object.Instantiate<GameObject>(femaleSkinBase);
		newCharacterContainer.SetActive(false);
		newCharacterContainer.name = futaSkinName;
		newCharacterContainer.transform.parent = femaleSkinsContainer.transform;
		
		var disableListener = newCharacterContainer.AddComponent<DisableListener>();
		disableListener.script = this;
		
		newCharacter = newCharacterContainer.GetComponentInChildren<DAZCharacter>();
		
		newCharacter.containingAtom = containingAtom;
		newCharacter.rootBonesForSkinning = characterSelector.rootBones;
		
		skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();	
		if(skin == null) {
			newCharacter.onLoadedHandlers = () => {
				ModifyAndMergeSkin();
			};	
		} else {
			ModifyAndMergeSkin();
		}


		// Set the new character as selected
		try {
		
			characterSelector.selectedCharacter = newCharacter;
		} catch (Exception e) {
			SuperController.LogError($"Error switching to new skin. {e}");
		}
	}

	public void ModifyAndMergeSkin() {
		skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();	

		// Skin should be loaded (but not initialized) by this point
		mergedMesh = (skin.dazMesh as DAZMergedMesh);
		skin.name = futaSkinName;

		// Load our mesh from file. This is the only way I found to initialize the
		// vertices manually. 
		// The file I'm using came from using StoreToBinaryFile, which required me to 
		// disable file IO restrictions. (Luckily LoadFromBinaryFile doesn't have those)
		genMesh = mergedMesh.gameObject.AddComponent<DAZMesh>();
		genMesh.LoadFromBinaryFile($"{GetPluginPath(this,containingAtom)}/male_gen_dazmesh");
		
		// The meshgraft includes instructions on which vertices should
		// be 'glued' to the base model. We're not bothering with this for now.
		// TODO: Set this...
		genMesh.meshGraft = maleMergedMesh.graftMesh.meshGraft;
		
		// Don't bother grafting the pelvis to the main body to prevent it deforming
		
		if(!hideMalePelvis.val){
			List<DAZMeshGraftVertexPair> graftPairs = new List<DAZMeshGraftVertexPair>(); 
			
			foreach(var pair in genMesh.meshGraft.vertexPairs){
				if(!RemovePins.Contains(pair.vertexNum) && !RemovePins.Contains(pair.graftToVertexNum)){
					graftPairs.Add(pair);
				}
			}
			
			var perineumPins = hideFemaleGens.val ? PerineumPins : PerineumPinsShowVagina;
			foreach(var entry in perineumPins){
				var pair = new DAZMeshGraftVertexPair();
				pair.vertexNum = entry.Key;
				pair.graftToVertexNum = entry.Value;
				graftPairs.Add(pair);
			}
			//graftPairs = new List<DAZMeshGraftVertexPair>();
			genMesh.meshGraft.vertexPairs = graftPairs.ToArray();
		
			var vagMesh = mergedMesh.graftMesh;
			
			
			var vaginaPins = hideFemaleGens.val ? FemaleGenPins : FemaleGenPinsShowVagina;
			List<DAZMeshGraftVertexPair> graftPairs2 = vagMesh.meshGraft.vertexPairs.ToList();
			foreach(var entry in vaginaPins){
				var pair = new DAZMeshGraftVertexPair();
				pair.vertexNum = entry.Key;
				pair.graftToVertexNum = entry.Value;
				graftPairs2.Add(pair);
			}
			//graftPairs2 = new List<DAZMeshGraftVertexPair>();
			vagMesh.meshGraft.vertexPairs = graftPairs2.ToArray();
			
			
			// Init gen mesh
			genMesh.meshGraft.hiddenPolys = new int[0];
			
			
			genMesh.graftTo = mergedMesh.targetMesh;
			
			genMesh.Init();
			
			// Remove penis vertices as needed
			DAZMeshData penisMeshData = ScriptableObject.CreateInstance(typeof(DAZMeshData)) as DAZMeshData;

			penisMeshData.baseVertices = genMesh.baseVertices;
			penisMeshData.basePolyList = (MeshPoly[])genMesh.basePolyList.Clone();
			penisMeshData.baseVerticesToUVVertices = genMesh.baseVerticesToUVVertices;
			penisMeshData.UVVertices = genMesh.UVVertices;
			penisMeshData.UVPolyList = (MeshPoly[])genMesh.UVPolyList.Clone();
			penisMeshData.OrigUV = genMesh.OrigUV;
			
			if(!hideFemaleGens.val){
				for(int i = 0; i < penisMeshData.UVPolyList.Length; i++){ 
					var meshPoly = penisMeshData.UVPolyList[i];
					
					foreach(int vert in meshPoly.vertices){
						if(HidePerineumVerts.Contains(vert)){
							meshPoly.materialNum = 2;
							break;
						}
					}
				}
			}
			genMesh.meshData = penisMeshData;	
			
			// Remove vagina vertices as needed
			DAZMeshData vagMeshData = ScriptableObject.CreateInstance(typeof(DAZMeshData)) as DAZMeshData;

			vagMeshData.baseVertices = vagMesh.baseVertices;
			vagMeshData.basePolyList = (MeshPoly[])vagMesh.basePolyList.Clone();
			vagMeshData.baseVerticesToUVVertices = vagMesh.baseVerticesToUVVertices;
			vagMeshData.UVVertices = vagMesh.UVVertices;
			vagMeshData.UVPolyList = (MeshPoly[])vagMesh.UVPolyList.Clone();
			vagMeshData.OrigUV = vagMesh.OrigUV;
			
			for(int i = 0; i < vagMeshData.UVPolyList.Length; i++){ 
				
				bool shouldAdd = !hideFemaleGens.val; // All verts on by default if showing vagina
				
				var meshPoly = vagMeshData.UVPolyList[i];
				foreach(int vert in meshPoly.vertices){
					if(!hideFemaleGens.val && HideFemalePelvisVertsWhenShowingVagina.Contains(vert)){
						meshPoly.materialNum = 3;
						break;
					}
					
					if(VisibleFemaleGenVerts.Contains(vert)){
						shouldAdd = true; 
						break;
					}
				}
				
				if(!shouldAdd){
					meshPoly.materialNum = 3;
				}
			}
			vagMesh.meshData = vagMeshData;
		
		
			// Remove the hip vertices
			// Typically this would be done by the hiddenPolys of the male gen graft,
			// but touching that seems to break clothing...
			DAZMeshData baseMeshData = ScriptableObject.CreateInstance(typeof(DAZMeshData)) as DAZMeshData;

			baseMeshData.baseVertices = mergedMesh.targetMesh.baseVertices;
			baseMeshData.basePolyList = (MeshPoly[])mergedMesh.targetMesh.basePolyList.Clone();
			baseMeshData.baseVerticesToUVVertices = mergedMesh.targetMesh.baseVerticesToUVVertices;
			baseMeshData.UVVertices = mergedMesh.targetMesh.UVVertices;
			baseMeshData.UVPolyList = (MeshPoly[])mergedMesh.targetMesh.UVPolyList.Clone();
			baseMeshData.OrigUV = mergedMesh.targetMesh.OrigUV;
			
			for(int i = 0; i < baseMeshData.UVPolyList.Length; i++){ 
				var meshPoly = baseMeshData.UVPolyList[i];
				foreach(int vert in meshPoly.vertices){
					if(HideHipVerts.Contains(vert)){
						meshPoly.materialNum = 31;
						break;
					}
				}
			}
			
			mergedMesh.targetMesh.meshData = baseMeshData;
		} else {
			genMesh.meshGraft.vertexPairs = new DAZMeshGraftVertexPair[0];
			genMesh.meshGraft.hiddenPolys = new int[0];
			genMesh.graftTo = mergedMesh.targetMesh;
			genMesh.Init();
		}
		
		// Merge the skin! 
		// This should result in the mergedMesh copying all of our gen mesh's 
		// vertices. 
		mergedMesh.Merge();
	
		
		// DAZSkinV2 also has a merge concept, which copies over the bone weights
		// that determine how the mesh is skinned to the skeleton.
		
		// For this we'll use the male skin we created earlier and copy its verts
		DAZSkinV2 maleGenSkin = null;
		foreach(var skinv2 in maleAsset.GetComponentsInChildren<DAZSkinV2>()){
			if(skinv2.dazMesh == maleMergedMesh.graftMesh){
				maleGenSkin = skinv2;
			}
		}
		
		DAZSkinV2Wrapper genSkinv2 = mergedMesh.gameObject.AddComponent<DAZSkinV2Wrapper>();
		genSkinv2.Copy(maleGenSkin);
		
		// Merge the skins.
		skin.Merge();
		skin.Init();
		
		// Adjust the male auto colliders
		foreach(var entry in maleAutoColliderConfigs){
			AutoCollider ac = entry.Key;
			AutoColliderConfig config = entry.Value;
			
			if(ac.targetVertex >=0) ac.targetVertex = config.targetVertex + mergedMesh.numGraftBaseVertices;
			if(ac.anchorVertex1 >=0) ac.anchorVertex1 = config.anchorVertex1 + mergedMesh.numGraftBaseVertices;
			if(ac.anchorVertex2 >=0) ac.anchorVertex2 = config.anchorVertex2 + mergedMesh.numGraftBaseVertices;
			if(ac.oppositeVertex >=0) ac.oppositeVertex = config.oppositeVertex + mergedMesh.numGraftBaseVertices;
		}
	
		// Set the materials
		
		Material[] originalMaterials = skin.GPUmaterials;
		skin.CopyMaterials();
		for (int i = 0; i < skin.dazMesh.materials.Length; i++)	{
			skin.materialsEnabled[i] = true;
			skin.dazMesh.materialsEnabled[i] = true;
		
		if(i < originalMaterials.Length){
				skin.GPUmaterials[i] = originalMaterials[i];
			} else {
				skin.GPUmaterials[i] = new Material(Shader.Find("Custom/Discard"));
			}
		}
		
		// Female gens
		if(useVaginaAsPelvisBase) {
			skin.GPUmaterials[31] = new Material(skin.GPUmaterials[28]);
		} else {
			skin.GPUmaterials[31] = new Material(skin.GPUmaterials[20]);
		}
		
		skin.FlushBuffers();
	}
		
	private IEnumerator WaitForLoad() {
		// Wait for it do its thing...
		while(newCharacterContainer == null || newCharacter == null || !newCharacter.ready){
			yield return null;
		}
		
		while(skin == null || !skin.wasInit){
			skin = newCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>();
			yield return null;
		}
		
		DAZMergedSkinV2 maleMergedSkin = maleCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>(true);
		while(maleMergedSkin == null){
			maleMergedSkin = maleCharacterContainer.GetComponentInChildren<DAZMergedSkinV2>(true);
			yield return null;
		}

		try{
			try {		
				// There are 3 morph banks for the (up to) 3 grafted meshes.
				// We want our 2nd graft (morphBank3) to use the morph's for 
				// The male's 1st graft (maleMorphBank2)
				characterRun.morphBank3 = characterSelector.maleMorphBank2;
				
				// morph banks will try to apply morphs to connected meshes
				// This might be set if the model was previously male. Need to unset it
				characterRun.morphBank3.connectedMesh = null;
				characterRun.ResetMorphs();	
				characterSelector.maleMorphBank2.gameObject.SetActive(true);
				SetMaleTransformsActive(true);
				
				SetUpMorphs();

				var useMaleMorphsOnFemale = characterSelector.GetBoolJSONParam("useMaleMorphsOnFemale");
				if(useMaleMorphsOnFemale != null && !useMaleMorphsOnFemale.val) {
					foreach(var morph in characterSelector.maleMorphBank1.morphs) {
						morph.morphValue = 0f;
					}
					useMaleMorphsOnFemale.val = true;
				}
				
			} catch (Exception e) {
				SuperController.LogError($"Error setting up positioning morphs. {e}");
			}

			// Penis
			skin.GPUmaterials[30] = new Material(maleMergedSkin.GPUmaterials[29]);
			if(useTransparentShader.val){
				skin.GPUmaterials[30].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
			} else {
				skin.GPUmaterials[30].shader = Shader.Find("Custom/Subsurface/GlossNMTessMappedFixedComputeBuff");
			}
			
			// Pelvis
			skin.GPUmaterials[29] = new Material(maleMergedSkin.GPUmaterials[28]);
			if(useTransparentShader.val){
				skin.GPUmaterials[29].shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
			} else {
				skin.GPUmaterials[29].shader = Shader.Find("Custom/Subsurface/GlossNMTessMappedFixedComputeBuff");
			}

			// Create the material options
			var defaultsMat = skin.GPUmaterials[29];
			if(useTransparentShader.val)
				defaultsMat.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff");
			

			genMaterialOptions = CreateMaterialOptions(new int[]{29,30}, defaultsMat);
			//pelvisMaterialOptions = CreateMaterialOptions(29, defaultsMat);
			//outerPelvisMaterialOptions = CreateMaterialOptions(31, outerPelvisDefault); //31

			RestoreSavedMaterialOptions();
			
			SyncAutoColliderSettings();
			
		} catch (Exception e){
			SuperController.LogMessage($"{e}");
		}
		
		// Re-enable the male portions of the model
		skinLoaded = true;
		
		// There's a race condition here that prevents the autocollider
		// update from taking effect, but I'm not sure what it is.
		// Waiting for a few frames seems to be enough. 
		yield return new WaitForSecondsRealtime(0.5f);
		
		UpdateAutoColliders();

		// I can't figure out a good way to force call AutoCollider.UpdateHardTransformPositionFast
		// Changing the morphs aught to do it though
		for(int i = 0; i < 10; i++) {
			genYOffsetMorph.morphValue += 0.001f;
			yield return new WaitForSecondsRealtime(0.1f);
			genYOffsetMorph.morphValue -= 0.001f;
		}
		
		
	}
	
	private void InitMainUI(){
		
		// Preset code was adapted from AcidBubbles' Collider Editor
		var loadPresetUI = CreateButton("Load Preset");
        loadPresetUI.button.onClick.AddListener(() =>
        {
            FileManagerSecure.CreateDirectory(altFutaSavePath);
            var shortcuts = FileManagerSecure.GetShortCutsForDirectory(altFutaSavePath);
            SuperController.singleton.GetMediaPathDialog(LoadPreset, altFutaFileExt, altFutaSavePath, false, true, false, null, false, shortcuts);
        });

        var savePresetUI = CreateButton("Save Preset");
        savePresetUI.button.onClick.AddListener(() =>
        {
            FileManagerSecure.CreateDirectory(altFutaSavePath);
            var fileBrowserUI = SuperController.singleton.fileBrowserUI;
            fileBrowserUI.SetTitle("Save AltFuta Preset");
            fileBrowserUI.fileRemovePrefix = null;
            fileBrowserUI.hideExtension = false;
            fileBrowserUI.keepOpen = false;
            fileBrowserUI.fileFormat = altFutaFileExt;
            fileBrowserUI.defaultPath = altFutaSavePath;
            fileBrowserUI.showDirs = true;
            fileBrowserUI.shortCuts = null;
            fileBrowserUI.browseVarFilesAsDirectories = false;
            fileBrowserUI.SetTextEntry(true);
            fileBrowserUI.Show(SavePreset);
            fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + altFutaFileExt;
            fileBrowserUI.ActivateFileNameField();
        });
		
		mergeLoad = new JSONStorableBool("Merge Load", false);
		RegisterBool(mergeLoad);
		var mergeLoadToggle = CreateToggle(mergeLoad, false);

		
		UIDynamicButton genMaterialOptionsButton = CreateButton("Penis Material Options", true);
		genMaterialOptionsButton.button.onClick.AddListener(() => {
			ShowMaterialOptionsUI(genMaterialOptionsUI);
		});

		UIDynamicButton reload = CreateButton("Reload Skin");
		reload.button.onClick.AddListener(() => {
			ReloadSkin();
		});
		
		CreateTextField(new JSONStorableString("hideFemaleGensText", "Don't render female gens, and instead renders the base G2F mesh. Does not have an anus."), false).height = 10;
		
		hideFemaleGens = new JSONStorableBool("Hide Female Gens", true);
		RegisterBool(hideFemaleGens);
		var hideGensToggle = CreateToggle(hideFemaleGens, false);
		hideGensToggle.toggle.onValueChanged.AddListener((val) => {
			ReloadSkin();
		});


		CreateTextField(new JSONStorableString("hideMalePelvisText", "Hides the pelvis portion of male gen mesh."), false).height = 10;
		hideMalePelvis = new JSONStorableBool("Hide Male Pelvis", false);
		RegisterBool(hideMalePelvis);
		var hideMalePelvisToggle = CreateToggle(hideMalePelvis, false);
		hideMalePelvisToggle.toggle.onValueChanged.AddListener((val) => {
			ReloadSkin();
		});

		CreateTextField(new JSONStorableString("transparentShaderText", "Use a transparent shader for the male gens. Allows for an alpha mask and blending, but doesn't support tesselation."), false).height = 10;
		useTransparentShader = new JSONStorableBool("Use Transparent Shader", false);
		RegisterBool(useTransparentShader);
		var transparentShaderToggle = CreateToggle(useTransparentShader, false);
		transparentShaderToggle.toggle.onValueChanged.AddListener((val) => {
			ReloadSkin();
		});

/*
		CreateTextField(new JSONStorableString("outerPelvisBase", "Which texture to use as the default for the female pelvis when hiding the female gens. Different ones work better for different skins."), false).height = 10;
		JSONStorableStringChooser outerPelvisSkinChooser = new JSONStorableStringChooser("outerPelvisBaseSkin", new List<string>{"Body", "Vagina"}, "Body", "Outer Pelvis Base Skin", (string baseSkin) => {
			useVaginaAsPelvisBase = baseSkin == "Vagina";
			overrideFemalePelvisTextures = true;
			ReloadSkin();
		});
		RegisterStringChooser(outerPelvisSkinChooser);
		
		CreateScrollablePopup(outerPelvisSkinChooser, false);
*/		
		genYOffsetValue = AddSlider("Bone+Mesh Y Offset", genBoneYOffsetDefault, -0.1f, 0.1f, (float value) => {
			if(genYOffsetMorph != null){
				genYOffsetMorph.morphValue = value;
			}
		});
		
		genZOffsetValue = AddSlider("Bone+Mesh Z Offset", genBoneZOffsetDefault, -0.1f, 0.1f, (float value) => {
			if(genZOffsetMorph != null){
				genZOffsetMorph.morphValue = value;
			}
		});
		
		genYMeshOffsetValue = AddSlider("Mesh Y Offset", 0f, -0.1f, 0.1f, (float value) => {
			if(genYMeshOffsetMorph != null){
				genYMeshOffsetMorph.morphValue = value;
			}
		});
		
		genZMeshOffsetValue = AddSlider("Mesh Z Offset", 0f, -0.1f, 0.1f, (float value) => {
			if(genZMeshOffsetMorph != null){
				genZMeshOffsetMorph.morphValue = value;
			}
		});
		
		hidePerineumOffsetValue = AddSlider("Hide Perimeum", hidePerineumDefault, -1f, 1f, (float value) => {
			if(hidePerineumMorph != null){
				hidePerineumMorph.morphValue = value;
			}
		});
		
		gapFixOffsetValue = AddSlider("Gap Fix", 0, 0f, 1f, (float value) => {
			if(gapFixMorph != null){
				gapFixMorph.morphValue = value;
			}
		});
		
		var vaginaHideDefault = hideFemaleGens.val? 1.0f : 0f;
		vaginaHideOffsetValue = AddSlider("Hide Vagina", vaginaHideDefault, 0f, 2f, (float value) => {
			if(vaginaHideMorph != null){
				vaginaHideMorph.morphValue = value;
			}
		});
		
		HideMaterialOptionsIUI();
	}
	
	public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, bool rightSide = false){
		JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
		storableFloat.storeType = JSONStorableParam.StoreType.Full;
		UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
		uiSlider.rangeAdjustEnabled = rangeAdjust;
		uiSlider.valueFormat = "F4";
		RegisterFloat(storableFloat);
		
		if(onChange != null){
			storableFloat.setCallbackFunction = onChange;
		}
		
		return storableFloat;
	}
	
	private IEnumerator InitMaterialUI(){
		// The material selction prefab doesn't do well with a layout,
		// so we're just going to remove it.
		
		while(scriptUI == null){
			if(this.UITransform != null){
				scriptUI = this.UITransform.GetComponentInChildren<MVRScriptUI>();
			}
			yield return null;
		}
		
		if(scriptUI == null){
			SuperController.LogMessage("Unable to load script UI.");
			yield break;
		}
		
		UnityEngine.UI.VerticalLayoutGroup layout = scriptUI.fullWidthUIContent.gameObject.GetComponent<UnityEngine.UI.VerticalLayoutGroup>();
		Destroy(layout);
		
		// Find a pre-existing material option UI to use as a template. 
		// Any would likely do...
		while( materialUIPrefab == null){
			var mouthMatTransform = containingAtom.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Mouth Materials");
			if(mouthMatTransform != null) {
				materialUIPrefab = mouthMatTransform.gameObject;
			} else {
				yield return null;
			}
		}
		
		// Wait for the skin to load
		while(skinLoaded == false){
			yield return null;
		}
		
		try {	
			if(genMaterialOptionsUI == null){
				genMaterialOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				genMaterialOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
			}
			genMaterialOptionsUI.gameObject.SetActive(false);
			genMaterialOptions.SetUI(genMaterialOptionsUI);
			genMaterialOptions.SyncAllParameters();
			genMaterialOptions.InitUI();
			
			
			/*if(pelvisMaterialOptionsUI == null){
				pelvisMaterialOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				pelvisMaterialOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
			}
			pelvisMaterialOptionsUI.gameObject.SetActive(false);
			pelvisMaterialOptions.SetUI(pelvisMaterialOptionsUI);
			pelvisMaterialOptions.SyncAllParameters();
			pelvisMaterialOptions.InitUI();
			*/
			
			/*if(outerPelvisMaterialOptionsUI == null){
				outerPelvisMaterialOptionsUI = UnityEngine.Object.Instantiate<Transform>(materialUIPrefab.transform);
				outerPelvisMaterialOptionsUI.SetParent(scriptUI.fullWidthUIContent,false);
			}
			outerPelvisMaterialOptionsUI.gameObject.SetActive(false);
			outerPelvisMaterialOptions.SetUI(outerPelvisMaterialOptionsUI);
			outerPelvisMaterialOptions.SyncAllParameters();
			outerPelvisMaterialOptions.InitUI();
			*/
			// Back button.
			// Not the prettiest but it'll do
			Transform backButtonTransform = UnityEngine.Object.Instantiate<Transform>(manager.configurableButtonPrefab.transform);
			backButtonTransform.SetParent(scriptUI.fullWidthUIContent,false);
			
			UIDynamicButton backButton = backButtonTransform.GetComponent<UIDynamicButton>();
			backButton.label = "BACK";
			backButton.buttonColor = Color.green;
			backButton.button.onClick.AddListener(() => {
				HideMaterialOptionsIUI();
			});
			
			RectTransform rect = backButtonTransform.gameObject.GetComponent<RectTransform>();
			rect.anchorMax = new Vector2(0.15f,1f);
			
			scriptUI.fullWidthUIContent.gameObject.SetActive(false);
			
			RestoreSavedMaterialOptions();
			
			
			
			overrideFemalePelvisTextures = false;
		} catch (Exception e) {
			SuperController.LogError($"Error initializing material options ui. {e}");
		}
	
	}
	
	private void SetUpMorphs(){
		var bank = characterSelector.maleMorphBank2;
		ZeroDuplicateMorphs(bank);
		
		genYOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Bone+Mesh Y-Offset.vmi");
		if(genYOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Bone+Mesh Y-Offset. Genitals may be positioned incorrectly.");
		} 
		
		genZOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Bone+Mesh Z-Offset.vmi");
		if(genZOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Bone+Mesh Z-Offset. Genitals may be positioned incorrectly.");
		}
		
		genYMeshOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Mesh Y-Offset.vmi");
		if(genYMeshOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Mesh Y-Offset. Genitals may be positioned incorrectly.");
		}
		
		genZMeshOffsetMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Mesh Z-Offset.vmi");
		if(genZMeshOffsetMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Mesh Z-Offset. Genitals may be positioned incorrectly.");
		}
		
		hidePerineumMorph = bank.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male_genitalia/AltFuta Hide Perineum.vmi");
		if(hidePerineumMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Hide Perineum. Genitals may be positioned incorrectly.");
		}
		
		gapFixMorph = characterSelector.maleMorphBank1.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/male/AltFuta Gap Fix.vmi");
		if(gapFixMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Gap Fix. Genitals may be positioned incorrectly.");
		}
		
		vaginaHideMorph = characterSelector.femaleMorphBank2.GetMorphByUid($"{packagePath}Custom/Atom/Person/Morphs/female_genitalia/AltFuta Vagina Hide.vmi");
		if(vaginaHideMorph == null){
			SuperController.LogError($"Unable to find morph AltFuta Vagina Hide. Genitals may be positioned incorrectly.");
		}

		genYOffsetValue.setCallbackFunction(genYOffsetValue.val);
		genZOffsetValue.setCallbackFunction(genZOffsetValue.val);
		genYMeshOffsetValue.setCallbackFunction(genYMeshOffsetValue.val);
		genZMeshOffsetValue.setCallbackFunction(genZMeshOffsetValue.val);
		hidePerineumOffsetValue.setCallbackFunction(hidePerineumOffsetValue.val);
		gapFixOffsetValue.setCallbackFunction(gapFixOffsetValue.val);
		vaginaHideOffsetValue.setCallbackFunction(vaginaHideOffsetValue.val);
	}
	
	
	// Adapted from Everlaster's TittyMagic v70
	private void ZeroDuplicateMorphs(DAZMorphBank morphBank)
	{
		if(string.IsNullOrEmpty(packageId))
		{
			return;
		}
		var versionRegex = new Regex(@"\.\d+$", RegexOptions.Compiled);
		
		string packageName = versionRegex.Split(packageId)[0];
		foreach(var morph in morphBank.morphs)
		{
			if(
				morph.isInPackage &&
				morph.morphValue != 0 &&
				morph.packageUid.StartsWith(packageName) &&
				morph.packageUid != packageId
			)
			{
				morph.morphValue = 0;
			}
		}
	}

	private void SetUpMeshMorph(DAZMorph morph, Vector3 direction){	
		morph.region = "Morph/Genitalia";
		morph.min = -1f;
		morph.max = 1f;
		morph.visible = true;
		morph.isRuntime = true;
		morph.isDemandLoaded = false;
		
		// Vertex Deltas
		var deltas = new List<DAZMorphVertex>();
		for(int i = 0; i < genMesh.baseVertices.Length; i++){
			deltas.Add(new DAZMorphVertex(){vertex = i, delta = direction});
		}
		morph.deltas = deltas.ToArray();	
		morph.numDeltas = morph.deltas.Length;
		
		morph.formulas = new DAZMorphFormula[0];
	}
	
	private void SetUpBoneMorph(DAZMorph morph, DAZMorphFormulaTargetType targetType, Vector3 direction){
		SetUpMeshMorph(morph,direction);

		// Bone Morphs
		var boneFormulas = new List<DAZMorphFormula>();
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen1", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen2", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Gen3", multiplier = 1f, targetType = targetType});
		boneFormulas.Add(new DAZMorphFormula(){target = "Testes", multiplier = 1f, targetType = targetType});
		
		morph.formulas = boneFormulas.ToArray();
	}
	
	private DAZMorphFormula MorphBoneFormula(string boneName, DAZMorphFormulaTargetType targetType){
		var formula = new DAZMorphFormula();
		formula.targetType = targetType;
		formula.target = boneName;
		formula.multiplier = 1f;
		return formula;
	}
	
	private MaterialOptionsWrapper CreateMaterialOptions(int[] matIndexes, Material defaultMaterial){
		var matOpts = newCharacterContainer.AddComponent<MaterialOptionsWrapper>();
		
		matOpts.textureGroup1 = new MaterialOptionTextureGroup();
		matOpts.SetCustomTextureFolder("Custom/Atom/Person/Textures");

		matOpts.materialForDefaults = defaultMaterial;

		matOpts.paramMaterialSlots = matIndexes;
		matOpts.textureGroup1.materialSlots = matOpts.paramMaterialSlots;
		
		matOpts.skin = skin;
		
		matOpts.enabled = true;
		matOpts.SyncAllParameters();
		return matOpts;
	}
	
	private void ShowMaterialOptionsUI(Transform matOptsUI){
		if(matOptsUI == null)
			return;
		
		if(genMaterialOptionsUI != null && genMaterialOptionsUI != matOptsUI)
			genMaterialOptionsUI.gameObject.SetActive(false);
		
		//if(pelvisMaterialOptionsUI != null && pelvisMaterialOptionsUI != matOptsUI)
		//	pelvisMaterialOptionsUI.gameObject.SetActive(false);
		
		//if(outerPelvisMaterialOptionsUI != null && outerPelvisMaterialOptionsUI != matOptsUI)
		//	outerPelvisMaterialOptionsUI.gameObject.SetActive(false);
		
		
		scriptUI.leftUIContent.gameObject.SetActive(false);
		scriptUI.rightUIContent.gameObject.SetActive(false);
		scriptUI.fullWidthUIContent.gameObject.SetActive(true);
		
		matOptsUI.gameObject.SetActive(true);
	}
	
	private void HideMaterialOptionsIUI(){
		if(genMaterialOptionsUI != null)
			genMaterialOptionsUI.gameObject.SetActive(false);
		
		//if(pelvisMaterialOptionsUI != null)
		//	pelvisMaterialOptionsUI.gameObject.SetActive(false);
		
		//if(outerPelvisMaterialOptionsUI != null)
		//	outerPelvisMaterialOptionsUI.gameObject.SetActive(false);
			
		
		if(scriptUI != null){
			scriptUI.leftUIContent.gameObject.SetActive(true);
			scriptUI.rightUIContent.gameObject.SetActive(true);
			scriptUI.fullWidthUIContent.gameObject.SetActive(false);
		}
	}
	
	private void RestoreSavedMaterialOptions(){
		try {
			if(genMaterialOptions == null){
				return;
			}
		
			// Restore the skin material options from the previous (non-futa) skin
			if(oldSkinOpts != null && newCharacter != null){
				foreach(var json in oldSkinOpts){
					foreach(var options in newCharacter.GetComponentsInChildren<DAZCharacterMaterialOptions>()){
						string id = json["id"];
						if(id != null && id == options.storeId){
							options.RestoreFromJSON(new JSONClass());
							options.RestoreFromJSON(json);
						}
					}
				}
				
			}
			
			var textureControl = newCharacter.GetComponentInChildren<DAZCharacterTextureControl>();
			if(textureControl != null && oldTextureControlJson != null){
				textureControl.RestoreFromJSON(new JSONClass());
				textureControl.RestoreFromJSON(oldTextureControlJson);
			}
			
			// Set the gen and pelivs material options to match the rest of the skin
			var skinMaterialOpts = containingAtom.GetStorableByID("skin") as DAZCharacterMaterialOptions;
		//	skinMaterialOpts.SyncAllParameters();
			genMaterialOptions.CopyMaterialParams(skinMaterialOpts);
			//pelvisMaterialOptions.CopyMaterialParams(skinMaterialOpts);
			//outerPelvisMaterialOptions.CopyMaterialParams(skinMaterialOpts);
			
			//if(!overrideFemalePelvisTextures){
			//	SyncFemalePelvisTextures(outerPelvisMaterialOptions, textureControl);
			//}
			
			// We don't want the textures in character skin options to apply to the male pelvis
			// which they do by default.
			genMaterialOptions.ClearCustomTextures();
			
			
			if(savedPresetJson != null){
				var genMatOptionsJson = savedPresetJson["genMaterialOpts"] as JSONClass;
				if(genMatOptionsJson != null && genMaterialOptions != null){
					genMaterialOptions.RestoreFromJSON(new JSONClass()); // Set all to default
					genMaterialOptions.RestoreFromJSON(genMatOptionsJson);
				}
				
				//var pelvisMatOptionsJson = savedPresetJson["pelvisMaterialOpts"] as JSONClass;
				//if(pelvisMatOptionsJson != null && pelvisMaterialOptions != null){
					
				//	pelvisMaterialOptions.RestoreFromJSON(new JSONClass()); // Set all to default
				//	pelvisMaterialOptions.RestoreFromJSON(pelvisMatOptionsJson);
				//}
				
				//var outerPelvisMatOptionsJson = savedPresetJson["outerPelvisMaterialOpts"] as JSONClass;
				//if(outerPelvisMatOptionsJson != null && outerPelvisMaterialOptions != null){
				//	outerPelvisMaterialOptions.RestoreFromJSON(outerPelvisMatOptionsJson);
				//}
			}
		} catch (Exception e) {
			SuperController.LogError($"Error copying skin options from previous skin. {e}");
		}
	}
	
	
	/*
		The texture control that controls the body's textures from the "Skin Textures" tab
		doesn't go through material options. It sets textures on the material directly. 
		
	*/
	private void SyncFemalePelvisTextures(MaterialOptions pelvisOpts, DAZCharacterTextureControl textureControl){
		var bodypart = "genitals";
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.textureName).val =  textureControl.GetUrlJSONParam(bodypart + "DiffuseUrl").val;
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.secondaryTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "SpecularUrl").val;
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.thirdTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "GlossUrl").val;
		//Fourth is alpha
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.fifthTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "NormalUrl").val;
		pelvisOpts.GetUrlJSONParam("customTexture" + pelvisOpts.textureGroup1.sixthTextureName).val =  textureControl.GetUrlJSONParam(bodypart + "DecalUrl").val;
		
	}
	
	private string GetPathToObject(GameObject go){
		 if (go.transform.parent == null || go.transform.parent == containingAtom.gameObject.transform)
			 return "/" + go.name;
		 return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
	}

	/* Change the status of the male portions of the model.
		Specifically we only want the rigid bodies and controllers
	*/
	private void SetMaleTransformsActive(bool active){
		foreach(GameObject go in maleObjects){
			go.SetActive(active);
		}
		
		foreach(var ac in genAutoColliders){
			ac.gameObject.SetActive(active);
		}
	}
	
	private void UpdateAutoColliders(){
		foreach( var entry in maleAutoColliderConfigs){
			var ac = entry.Key;
			var resizeTrigger = ac.resizeTrigger;
			ac.resizeTrigger = AutoCollider.ResizeTrigger.Always;
			ac.AutoColliderSizeSet(true);
			ac.resizeTrigger = resizeTrigger;
			
			ac.resetSimulation = true;
		}
	}
	
	private void SyncAutoColliderSettings(){
		foreach(GameObject go in maleObjects){
			if(go.activeInHierarchy){
				foreach(AutoCollider ac in genAutoColliderIgnoreList){
					var ignoreList = ac.ignoreColliders.ToList();
					if(!ignoreList.Contains(go.transform)){
						ignoreList.Add(go.transform);
						ac.ignoreColliders = ignoreList.ToArray();
						ac.InitColliders();
					}
				}
			}
		}
		
		foreach(var maleAutoCollider in genAutoColliders){
			if(maleAutoCollider.hardCollider != null){
				
				
				// Gen colliders shouldn't collide with each other...
				foreach(var maleAutoCollider2 in genAutoColliders){
					if(maleAutoCollider != maleAutoCollider2){
						Physics.IgnoreCollision(maleAutoCollider.hardCollider, maleAutoCollider2.hardCollider);
					}
				}
				
				
			}
			IgnoreCollider(maleAutoCollider.hardCollider);
			IgnoreCollider(maleAutoCollider.jointCollider);
		}
		
		foreach(var maleCollider in genColliders){
			IgnoreCollider(maleCollider);
		}
		
		foreach(var batchUpdater in characterRun.autoColliderUpdaters){
			batchUpdater.UpdateAutoColliders();
		}
		
		
	}
	
	private void IgnoreCollider(Collider collider){
		if(collider == null){
			return;
		}
		
		foreach(AutoCollider ac in genAutoColliderIgnoreList){
			if(ac.jointCollider != null){
				Physics.IgnoreCollision(collider, ac.jointCollider);
			}
		}
		
		foreach(Collider c in genColliderIgnoreList){
			//SuperController.LogMessage($"{collider} {c}");
			Physics.IgnoreCollision(collider, c);
		}
		
		var mesh = containingAtom.GetStorableByID("LowerPhysicsMesh") as DAZPhysicsMesh;
		foreach(DAZPhysicsMeshSoftVerticesGroup vertGroup in mesh.softVerticesGroups){
			foreach(var softVertSet in vertGroup.softVerticesSets){
				Physics.IgnoreCollision(softVertSet.jointCollider, collider);
				
				if (softVertSet.jointCollider2 != null){
					Physics.IgnoreCollision(softVertSet.jointCollider2, collider);
				}
			}
		}
	}
	
	private void Update(){
		// For whatever reason, putting on clothes resets this;
		if( skin != null){
			skin.materialsEnabled[29] = !hideMalePelvis.val;
			//skin.materialsEnabled[28]= !hideFemaleGens.val;
			//skin.materialsEnabled[28]= false;
			skin.materialsEnabled[31] = false;
		}
		
		if(characterRun.morphBank3 != null){
			// Fix these settings in case they got changed somehow
			if(characterRun.morphBank3 != characterSelector.maleMorphBank2){
				characterRun.morphBank3 = characterSelector.maleMorphBank2;
			}

			if(characterRun.morphBank3.connectedMesh != null){
				characterRun.morphBank3.connectedMesh = null;
			}
			
			
			// For whatever reason, the threaded morph logic in DAZCharacterRun
			// just doesn't bother with the 2nd graft. 
			// So we'll just do it manually here
			
			if ( characterRun.morphBank3.bonesDirty)	{
				characterRun.WaitForRunTask();
				characterRun.bones.SetMorphedTransform(false);
				characterRun.morphBank3.bonesDirty = false;
			}
			if (characterRun.morphBank3.boneRotationsDirty != null && characterRun.morphBank3.boneRotationsDirty.Count > 0) {
				characterRun.WaitForRunTask();
				foreach (var entry in characterRun.morphBank3.boneRotationsDirty){
					entry.Key.SyncMorphBoneRotations(false);
				}
				characterRun.morphBank3.boneRotationsDirty.Clear();
			}
		}
		
		//if (Input.GetMouseButtonDown(1)){
		//}
	}
	
	
	
	private void UnloadSkin(){
		skin = null;
		genMaterialOptions = null;
		//pelvisMaterialOptions = null;
		//outerPelvisMaterialOptions = null;
		
		if(maleCharacterContainer != null){
			Destroy(maleCharacterContainer);
		}
		if(newCharacterContainer != null){
			Destroy(newCharacterContainer);
		}
		if(maleAsset != null) {
			Destroy(maleAsset);
		}
		
		maleCharacterContainer = null;
		newCharacterContainer = null;
		maleAsset = null;
	}
	
	private void OnEnable(){
		SetMaleTransformsActive(true);
		ReloadSkin();
	}
	
	private void OnDisable(){
		SetMaleTransformsActive(false);
		if(characterSelector != null && previousSelected != null && characterSelector.selectedCharacter == newCharacter){
			characterSelector.selectedCharacter = previousSelected;
		}
	}
	
	private void OnDestroy()
	{	
		UnloadSkin();
	}
		
	List<GameObject> GetAllChildren(GameObject parent, List<GameObject> objectList = null)
	{
		if (objectList == null) objectList = new List<GameObject>();
	  
		foreach (Transform child in parent.transform) {
			objectList.Add(child.gameObject);
			GetAllChildren(child.gameObject, objectList);
		}
		return objectList;
	}
	
	public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
		savedPresetJson = jc;
		
		if(jc["baseSkinMaterialOpts"] != null){
			var optsList = new List<JSONClass>();
			foreach(var optsJson in jc["baseSkinMaterialOpts"].AsArray){
				optsList.Add(optsJson as JSONClass);
			}
			oldSkinOpts = optsList.ToArray();
		}
		
		if(jc["baseTextureControl"] != null){
			oldTextureControlJson = jc["baseTextureControl"].AsObject;
		}
		
		base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
	}
	
	public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
	{
		var jsonClass = base.GetJSON(true, true, true);
		
		jsonClass.Add("genMaterialOpts", genMaterialOptions.GetJSON(true,true,true)); 
		
		//jsonClass.Add("pelvisMaterialOpts", pelvisMaterialOptions.GetJSON(true,true,true)); 
		
		//jsonClass.Add("outerPelvisMaterialOpts", outerPelvisMaterialOptions.GetJSON(true,true,true)); 
		
		JSONArray baseSkinMaterialOpts = new JSONArray();
		foreach(var options in newCharacter.GetComponentsInChildren<DAZCharacterMaterialOptions>()){
			baseSkinMaterialOpts.Add(options.GetJSON(true,true,true));
		}		
		jsonClass.Add("baseSkinMaterialOpts", baseSkinMaterialOpts); 
		
		
		var textureControl = newCharacter.GetComponentInChildren<DAZCharacterTextureControl>();
		jsonClass.Add("baseTextureControl", textureControl.GetJSON(true,true,true)); 
		
		return jsonClass;
	}
	
	private void LoadPreset(string path) {
        if (string.IsNullOrEmpty(path))
            return;
		
		var presetJson = (JSONClass)LoadJSON(path);
		
		// Adjust the material options first. 
		savedPresetJson = presetJson;
		
		if(!mergeLoad.val){
			if(savedPresetJson["baseSkinMaterialOpts"] != null){
				var optsList = new List<JSONClass>();
				foreach(var optsJson in savedPresetJson["baseSkinMaterialOpts"].AsArray){
					optsList.Add(optsJson as JSONClass);
				}
				oldSkinOpts = optsList.ToArray();
			}
			
			if(savedPresetJson["baseTextureControl"] != null){
				oldTextureControlJson = savedPresetJson["baseTextureControl"].AsObject;
			}
		} else {
			oldSkinOpts = null;
			oldTextureControlJson = null;
		}
		
		RestoreSavedMaterialOptions();


		// Restore the other settings. This might trigger a reload
        RestoreFromJSON(presetJson, true, true, null, false);
		
		
    }
	
	private void SavePreset(string path) {
        SuperController.singleton.fileBrowserUI.fileFormat = null;
        if (string.IsNullOrEmpty(path)) {
			return;
		}

        if (!path.ToLower().EndsWith($".{altFutaFileExt}")) {
            path += $".{altFutaFileExt}";
		}

        SaveJSON(GetJSON(true,true,true), path);
    }
	
	// Function taken from VAMDeluxe's code
	public static string GetPluginPath(MVRScript plugin, Atom atom){
		SuperController.singleton.currentSaveDir = SuperController.singleton.currentLoadDir;
		string pluginId = plugin.storeId.Split('_')[0];
		MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
		string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
		string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
		return pathToScriptFolder;
	}
	
	
	//MacGruber / Discord 20.10.2020
	//Get path prefix of the package that contains this plugin
	private static string GetPackageId(MVRScript script)
	{
		string id = script.name.Substring(0, script.name.IndexOf('_'));
		string filename = script.manager.GetJSON()["plugins"][id].Value;
		int idx = filename.IndexOf(":/", StringComparison.Ordinal);
		return idx >= 0 ? filename.Substring(0, idx) : null;
	}
	
	// TODO: Move these to a separate file maybe...
	List<int> RemovePins = new List<int>(){
		1190,1188,1198,1189,
		1200,1186,1192,1119,
		1120,1113,1129,1116,
		1117,1115,1127,1104,
		1105,1128,1103,1130,
		1102,1121,1193,1201,
		1178,1199,1179,1180,
		1177,1122,1107,1233,
		6122,16844
	};
	
	List<int> HideHipVerts = new List<int>(){
		18583,18600,18597,18596,
		18840,18836,18848,18837,
		18839,18838,8164,8165,
		8166,8163,8176,8162,
		8167,7918,7919,7922,
		7905,
	};
	
	Dictionary<int,int> FemaleGenPins = new Dictionary<int,int>{
						
		//test
		[159] = 2781,
		//[158] = 8194,
		[1348] = 7816,

		//[649] = 8186,
		[649] = 7859,
		[279] = 8196,
		//[674] = 8184,					
		[674] = 2778,
					
		// test r
		[756] = 13589,
		[1343] = 18496,

		[1349] = 18496,
		//[1240] = 18857,
		[1240] = 18539,
		[885] = 18867,
								
		// Double-layer Left
		[33] = 8215,
		[4] = 2326,
		[2] = 2311,
		[7] = 6122,
		[6] = 6121,
		[0] = 334,
		[5] = 2777,		


		// Double-layer right
		[67] = 18886,
		[39] = 13156,
		[37] = 13141,
		[42] = 16844,
		[41] = 16843,
		[36] = 11262,
		[40] = 13586,				
	};
	
	Dictionary<int,int> FemaleGenPinsShowVagina = new Dictionary<int,int>{
						

								
		// Double-layer Left
		[33] = 8215,
		[4] = 2326,
		[2] = 2311,
		[7] = 6122,
		[6] = 6121,
		[0] = 334,
		[5] = 2777,		

		// Double-layer right
		[67] = 18886,
		[39] = 13156,
		[37] = 13141,
		[42] = 16844,
		[41] = 16843,
		[36] = 11262,
		[40] = 13586,

		// Extra
		[1296] = 7775,
		[116] = 7817,
		[1298] = 18455,
		[714] = 18497,
		
		//[1311] = 13142,
		//[1308] = 2312,
		
		// scrotum
		/*[1309] = 2313,
		[125] = 2314,
		[103] = 6766,
		[1310] = 13143,
		[723] = 13144,
		[701] = 17475,*/
[701] = 13144,
[103] = 2314,
[732] = 17475,
[134] = 6766,

[125] = 2313,
[1309] = 2312,
[1310] = 13142,
[723] = 13143,

// clit


	};
	
	Dictionary<int,int> PerineumPins = new Dictionary<int,int>{
		// Top Row
	/*	[1202] = 18583,
		[1212] = 18600,
		[1210] = 18597,
		[1209] = 18596,
		[1222] = 18840,
		[1218] = 18836,
		[1230] = 18848,
		[1219] = 18837,
		[1221] = 18839,
		[1220] = 18838,
		[1150] = 8164,
		[1151] = 8165,
		[1152] = 8166,
		[1149] = 8163,
		[1162] = 8176,
		[1148] = 8162,
		[1153] = 8167,
		[1138] = 7918,
		[1139] = 7919,
		[1141] = 7922,
		[1131] = 7905,*/

		//Second Row

	/*	[1214] = 18621,
		[1206] = 18591,
		[1185] = 13750,
		[1227] = 18845,
		[1228] = 18846,
		[1229] = 18847,
		[1231] = 18849,
		[1216] = 18834,
		[1217] = 18835,
		[1215] = 18833,
		[1144] = 8158,
		[1145] = 8159,
		[1147] = 8161,
		[1146] = 8160,
		[1163] = 8177,
		[1161] = 8175,
		[1160] = 8174,
		[1159] = 8173,
		[1112] = 2956,
		[1135] = 7913,
		[1143] = 7943,*/

		// Left 
		[1118] = 6122,
		[1117] = 6121,
		[1104] = 334,
			
		// 
		[1105] = 2777,				
		[1128] = 2781,
		[1103] = 7816,

	//	[1130] = 8186,
		[1130] = 7859,
		[1102] = 8196,	
	//	[1107] = 8184,
		[1107] = 2778,

		//
		[1191] = 16844,
		[1190] = 16843,
		[1179] = 11262,

		[1180] = 13586,
		[1199] = 13589,
		[1178] = 18496,

		//[1201] = 18857,
		[1201] = 18539,
		
		[1177] = 18867,	
	};
	
	Dictionary<int,int> PerineumPinsShowVagina = new Dictionary<int,int>{
		// Top Row
	/*	[1202] = 18583,
		[1212] = 18600,
		[1210] = 18597,
		[1209] = 18596,
		[1222] = 18840,
		[1218] = 18836,
		[1230] = 18848,
		[1219] = 18837,
		[1221] = 18839,
		[1220] = 18838,
		[1150] = 8164,
		[1151] = 8165,
		[1152] = 8166,
		[1149] = 8163,
		[1162] = 8176,
		[1148] = 8162,
		[1153] = 8167,
		[1138] = 7918,
		[1139] = 7919,
		[1141] = 7922,
		[1131] = 7905,*/

		//Second Row

	/*	[1214] = 18621,
		[1206] = 18591,
		[1185] = 13750,
		[1227] = 18845,
		[1228] = 18846,
		[1229] = 18847,
		[1231] = 18849,
		[1216] = 18834,
		[1217] = 18835,
		[1215] = 18833,
		[1144] = 8158,
		[1145] = 8159,
		[1147] = 8161,
		[1146] = 8160,
		[1163] = 8177,
		[1161] = 8175,
		[1160] = 8174,
		[1159] = 8173,
		[1112] = 2956,
		[1135] = 7913,
		[1143] = 7943,*/

		// inner thighs
		[1194] = 18455,
		[1256] = 18497,
		[1123] = 7775,
		[1255] = 7817,
	

//		[307] = 2312,
//		[743] = 13142,
	
		// scrotum
		/*[309] = 2313,
		[312] = 2314,
		[310] = 6766,
		
		[311] = 6765,
		
		[745] = 13143,
		[747] = 13144,
		[746] = 17475,*/
		
		
[311] = 6765,
[310] = 6766,
[746] = 17475,
[312] = 2314,
[747] = 13144,
[309] = 2313,
[745] = 13143,
[307] = 2312,
[743] = 13142,
	};
					

	
	List<int> VisibleFemaleGenVerts = new List<int>(){
		1431,
		1423,
		1429,
		1430,
		1426,
		1415,
		1414,
		1413,
		1418,
		1417,
		1412,
		1416,
		1425,
		1424,
		1451,
		1443,
		1449,
		1450,
		1446,
		1435,
		1434,
		1433,
		1438,
		1437,
		1432,
		1436,
		1445,
		1444,		

		// back line
		8187,
		8252,
		7189,
		7840,
		7868,
		7820,
		7868,
		7854,
		7874,
		18554,
		18534,
		18548,
		18500,
		18520,
		17896,
		18858,
		18923,

		// second line
		8197,
		8248,
		11,
		1261,
		1301,
		155,
		156,
		1300,
		1297,
		1299,
		753,
		752,
		1302,
		1263,
		46,
		18919,
		18868,

		12,
		9,
		1323,
		1303,
		1341,
		1339,
		1337,
		1338,
		1340,
		1304,
		1324,
		44,
		47,
		43,
		18866,
		8195,
		8,

		2777,
		8195,
		157,
		1319,
		644,
		646,
		1330,
		647,
		290,
		143,
		893,
		1238,
		1334,
		1237,
		1235,
		1321,
		754,


		8194,
		1320,
		648,
		645,
		151,
		152,
		1329,
		153,
		154,
		75,
		751,
		750,
		1331,
		749,
		748,
		1236,
		1239,
		1322,
		755,
		741,
		145,
		1350,
		146,
		147,
		1325,
		148,
		149,
		144,
		746,
		745,
		1332,
		744,
		743,
		1358,
		742,
		1345,
		1356,
		1355,
		1354,
		150,
		676,
		747,
		1357,
		1360,
		1361,
		1344,
		1234,
		1233,
		721,
		675,
		123,
		642,
		643,

		// ?


		158,


		673,
		528,
		1328,
		527,
		526,
		1353,
		525,
		1353,
		525,
		1113,
		1114,
		1363,
		1115,
		1116,
		1335,
		1116,
		1335,
		1116,
		1335,
		1116,
		1336,
		1117,
		1118,
		1119,
		529,
		528,

		672,
		295,
		294,
		1327,
		292,
		293,
		292,
		1352,
		291,
		79,
		894,
		1362,
		895,
		896,
		901,
		896,
		901,
		1335,
		897,
		898,
	};
	
	List<int> HideFemalePelvisVertsWhenShowingVagina = new List<int>(){
		1439,
		56,
		55,
		54,
		63,
		59,
		64,
		60,
		62,
		61,
		26,
		27,
		28,
		25,
		30,
		24,
		29,
		19,
		20,
		21,
		1419,
		1442,
		53,
		1262,
		1276,
		1275,
		678,
		65,
		58,
		677,
		57,
		70,
		22,
		80,
		23,
		31,
		1259,
		1272,
		1271,
		1260,
		18,
		1422,

		1284,
		1294,
		1295,
		1282,
		1290,
		1291,
		1283,
		1286,
		1287,
		1285,
		1265,
		1266,
		1264,
		1279,
		1288,
		1289,
		1280,
		1292,
		1293,
		1281,
		681,
		1277,
		1278,
		680,
		1269,
		1270,
		679,
		71,
		81,
		1267,
		1268,
		82,
		1273,
		83,
		1274,
		684,
		683,
		682,
		72,
		84,
		85,
		86,
		687,
		686,
		685,
		665,
		87,
		88,
		89,
		
		// fix
		160,
		757,
	};
	
	

	List<int> HidePerineumVerts = new List<int>(){
		1104,
		1105,
		1128,
		1130,
		1103,
		1102,
		1107,
		1177,
		1201,
		1178,
		1199,
		1180,
		1179,
		1180,
		1199,
		1193,
		1122,
		1121,
		
		1117,
		1115,
		1127,
		1116,
		1129,
		//1113,
		//1120,
		//1119,
		//1192,
		//1186,
		1200,
		1189,
		1198,
		1188,
		1190,
		
		1118,
		1114,
		1126,
		1264,
		//1262,
		//1260,
		//1258,
		//1257,
		//1259,
		//1261,
		//1263,
		1265,
		1197,
		1187,
		1191,
	};
} // script
} // namespace 